@@ -1,6 +1,10 @@
+import struct
 import sys
-from datetime import datetime
+from datetime import datetime, timezone
+from decimal import Decimal
 
+from cryptography.exceptions import InvalidSignature
+from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey
 from tortoise.fields import (
     SmallIntField,
     IntField,
@@ -10,7 +14,6 @@ from tortoise.fields import (
     IntEnumField,
     FloatField,
     JSONField,
-    BinaryField,
     ForeignKeyField,
     OneToOneField,
     ManyToManyField,
@@ -19,8 +22,10 @@ from tortoise.fields import (
     ForeignKeyNullableRelation,
     OneToOneNullableRelation,
     ManyToManyRelation,
+    BinaryField,
 )
 from tortoise.fields.relational import BackwardFKRelation, BackwardOneToOneRelation
+from tortoise.functions import Sum
 from tortoise.queryset import QuerySet
 from tortoise import Model as BaseModel
 from tortoise.signals import pre_save
@@ -60,12 +65,17 @@ from xync_schema.enums import (
     AbuserType,
     Boundary,
     SbpStrict,
+    TransactionStatus,
 )
 
 
+class UniqBinaryField(BinaryField):
+    indexable = True
+
+
 class Country(BaseCountry):
     cur: ForeignKeyRelation["Cur"] = ForeignKeyField("models.Cur", related_name="countries")
-    curexs: ManyToManyRelation["Curex"]
+    curexs: ManyToManyRelation["CurEx"]
     forbidden_exs: ManyToManyRelation["Ex"]
     fiats: BackwardFKRelation["Fiat"]
 
@@ -91,9 +101,9 @@ class Cur(Model):
 class Coin(Model):
     id: int = SmallIntField(True)
     ticker: str = CharField(15, unique=True)
-    scale: int = SmallIntField(default=0)
     rate: float | None = FloatField(default=0)
     is_fiat: bool = BooleanField(default=False)
+    scale: int = SmallIntField(default=4)
     exs: ManyToManyRelation["Ex"] = ManyToManyField("models.Ex", through="coinex")
 
     assets: BackwardFKRelation["Asset"]
@@ -119,19 +129,19 @@ class Ex(Model):
     host: str | None = CharField(63, null=True, description="With no protocol 'https://'")
     host_p2p: str | None = CharField(63, null=True, description="With no protocol 'https://'")
     url_login: str | None = CharField(63, null=True, description="With no protocol 'https://'")
-    type_: ExType = IntEnumField(ExType)
+    typ: ExType = IntEnumField(ExType)
     status: ExStatus = IntEnumField(ExStatus, default=ExStatus.plan)
     logo: str = CharField(511, default="")
 
     ads: ManyToManyRelation["Ad"]
     pms: ManyToManyRelation["Pm"]
     curs: ManyToManyRelation[Cur]
-    # pmcurs: ManyToManyRelation["Pmcur"] = ManyToManyField("models.Pmcur", through="pmcurex")
+    # pmcurs: ManyToManyRelation["PmCur"] = ManyToManyField("models.PmCur", through="pmcurex")
     coins: ManyToManyRelation[Coin]
     forbidden_countries: ManyToManyRelation[Country] = ManyToManyField("models.Country", related_name="forbidden_exs")
 
     actors: BackwardFKRelation["Actor"]
-    pmexs: BackwardFKRelation["Pmex"]
+    pmexs: BackwardFKRelation["PmEx"]
     pm_reps: BackwardFKRelation["PmRep"]
     pairexs: BackwardFKRelation["PairEx"]
     deps: BackwardFKRelation["Dep"]
@@ -139,40 +149,41 @@ class Ex(Model):
 
     class Meta:
         table_description = "Exchanges"
-        unique_together = (("name", "type_"),)
+        unique_together = (("name", "typ"),)
 
     class PydanticMeta(Model.PydanticMeta):
         include = "name", "logo"
 
-    def client(self, bot, actor: "Actor" = None, **kwargs):
+    def client(self, bot, **kwargs):
         module_name = f"xync_client.{self.name}.ex"
         __import__(module_name)
         client = sys.modules[module_name].ExClient
         return client(self, bot, **kwargs)
 
 
-class Curex(BaseModel):
+class CurEx(BaseModel):
     cur: ForeignKeyRelation[Cur] = ForeignKeyField("models.Cur")
     cur_id: int
     ex: ForeignKeyRelation[Ex] = ForeignKeyField("models.Ex")
     ex_id: int
     exid: str = CharField(32)
-    minimum: int = IntField(null=True)
-    rounding_scale: int = SmallIntField(null=True)
-    countries: ManyToManyRelation[Country] = ManyToManyField(
-        "models.Country", through="curexcountry", backward_key="curexs"
-    )
+    minimum: int = IntField(null=True)  # /10^cur.scale
+    scale: int = SmallIntField(null=True)
+    # countries: ManyToManyRelation[Country] = ManyToManyField (
+    #     "models.Country", through="curex_country", backward_key="curexs"
+    # )
 
     class Meta:
         table_description = "Currency in Exchange"
         unique_together = (("ex_id", "cur_id"), ("ex_id", "exid"))
 
 
-class Coinex(BaseModel):
+class CoinEx(BaseModel):
     coin: ForeignKeyRelation[Coin] = ForeignKeyField("models.Coin")
     coin_id: int
     ex: ForeignKeyRelation[Ex] = ForeignKeyField("models.Ex")
-    minimum: float = FloatField(null=True)
+    minimum: int = IntField(null=True)  # /10^self.scale
+    scale: int = SmallIntField(default=4)
 
     exid: str = CharField(32)
     p2p: bool = BooleanField(default=True)
@@ -220,10 +231,10 @@ class NetAddr(Model):  # Way
 
 
 class PairEx(Model, TsTrait):  # todo: refact to PairSideEx?
-    # todo: различаются ли комишки на buy/sell по одной паре хоть на одной бирже? если да, то переделать
+    # todo: различаются ли комиссии на buy/sell по одной паре хоть на одной бирже? если да, то переделать
     pair: ForeignKeyRelation[Pair] = ForeignKeyField("models.Pair", related_name="pairexs")
     pair_id: int
-    fee: float = FloatField(default=0)
+    fee: int = SmallIntField(default=0)  # /10_000
     ex: ForeignKeyRelation[Ex] = ForeignKeyField("models.Ex", related_name="pairs")
     ex_id: int
     pair_sides: BackwardFKRelation["PairSide"]
@@ -255,10 +266,8 @@ class User(TgUser, TsTrait):
     person_id: int
     ref: ForeignKeyNullableRelation["User"] = ForeignKeyField("models.User", related_name="proteges", null=True)
     ref_id: int | None
-    contacted_with: ForeignKeyNullableRelation["User"] = ForeignKeyField(
-        "models.User", related_name="contacts", null=True
-    )  # who can texts this user
-    contacted_with_id: int | None
+    prv = BinaryField(null=True)  # len=32
+    pub = UniqBinaryField(unique=True, null=True)  # len=32
 
     actors: BackwardFKRelation["Actor"]
     contacts: BackwardFKRelation["User"]
@@ -270,13 +279,15 @@ class User(TgUser, TsTrait):
     limits: BackwardFKRelation["Limit"]
     msgs: BackwardFKRelation["Msg"]
     proteges: BackwardFKRelation["User"]
-
-    # vpn: BackwardOneToOneRelation["Vpn"]
-    # invite_requests: BackwardFKRelation["Invite"]
-    # invite_approvals: BackwardFKRelation["Invite"]
-    # lends: BackwardFKRelation["Credit"]
-    # borrows: BackwardFKRelation["Credit"]
-    # investments: BackwardFKRelation["Investment"]
+    sends: BackwardFKRelation["Transaction"]
+    receives: BackwardFKRelation["Transaction"]
+    validates: BackwardFKRelation["Transaction"]
+    vpn: BackwardOneToOneRelation["Vpn"]
+    invite_requests: BackwardFKRelation["Invite"]
+    invite_approvals: BackwardFKRelation["Invite"]
+    lends: BackwardFKRelation["Credit"]
+    borrows: BackwardFKRelation["Credit"]
+    investments: BackwardFKRelation["Investment"]
 
     async def free_assets(self):
         assets = await Asset.filter(agent__actor__person__user__id=self.id).values("free", "addr__coin__rate")
@@ -284,10 +295,123 @@ class User(TgUser, TsTrait):
 
     async def fiats_sum(self):
         fiats = await Fiat.filter(cred__person__user__id=self.id).values("amount", "cred__pmcur__cur__rate")
-        return sum(fiat["amount"] * fiat["cred__pmcur__cur__rate"] for fiat in fiats)
+        return sum(fiat["amount"] * Decimal(fiat["cred__pmcur__cur__rate"]) for fiat in fiats)
 
-    async def balance(self) -> float:
-        return await self.free_assets() + await self.fiats_sum()
+    async def balance_sum(self) -> int:
+        return int(await self.free_assets()) + int(await self.fiats_sum())
+
+    async def balance_xp(self, cur_id: int = None) -> dict[int, int] | int:
+        if cur_id:
+            dbt = (
+                await Transaction.filter(receiver=self, cur_id=cur_id)  # , status=TransactionStatus.valid
+                .group_by("cur_id")
+                .annotate(debit=Sum("amount"))
+                .values_list("debit", flat=True)
+            ) or [0]
+            crd = (
+                await Transaction.filter(sender=self, cur_id=cur_id)  # , status=TransactionStatus.valid
+                .group_by("cur_id")
+                .annotate(credit=Sum("amount"))
+                .values_list("credit", flat=True)
+            ) or [0]
+            return dbt[0] - crd[0]
+        dbt = {
+            c: v
+            for c, v in await Transaction.filter(receiver=self, status=TransactionStatus.valid)
+            .group_by("cur_id")
+            .annotate(debit=Sum("amount"))
+            .values_list("cur_id", "debit")
+        }
+        crd = {
+            c: v
+            for c, v in await Transaction.filter(sender=self, status=TransactionStatus.valid)
+            .group_by("cur_id")
+            .annotate(credit=Sum("amount"))
+            .values_list("cur_id", "credit")
+        }
+        return {c: dbt.get(c, 0) - crd.get(c, 0) for c in dbt.keys() | crd.keys()}
+
+    async def keygen(self):
+        prv = Ed25519PrivateKey.generate()
+        self.prv = prv.private_bytes_raw()
+        self.pub = prv.public_key().public_bytes_raw()
+        await self.save(update_fields=["prv", "pub"])
+
+    async def get_val_id(self, amount: int, cur_id: int, receiver_id: int) -> int:
+        r = {
+            c: v
+            for c, v in await Transaction.filter(
+                receiver_id__not_in=[self.id, receiver_id], status=TransactionStatus.valid, cur_id=cur_id
+            )
+            .group_by("receiver_id")
+            .annotate(debit=Sum("amount"))
+            .values_list("receiver_id", "debit")
+        }
+        s = {
+            c: v
+            for c, v in await Transaction.filter(
+                sender_id__not_in=[self.id, receiver_id], status=TransactionStatus.valid, cur_id=cur_id
+            )
+            .group_by("sender_id")
+            .annotate(credit=Sum("amount"))
+            .values_list("sender_id", "credit")
+        }
+        # noinspection PyUnboundLocalVariable
+        vals = {c: am for c in r.keys() | s.keys() if c and (am := r.get(c, 0) - s.get(c, 0)) > amount}
+        val_id, _ = sorted(vals.items(), key=lambda x: x[1], reverse=True)[0]
+        if not val_id:
+            raise ValueError("No available validators:(")
+        validator = await User[val_id]
+        if not validator.pub or not validator.prv or not len(validator.pub + validator.prv) == 64:  # todo: remove check
+            await validator.keygen()
+        return val_id
+
+    async def req(self, amount: int, cur_id: int, sender_id: int = None) -> "Transaction":
+        return await Transaction.create(
+            receiver=self,
+            sender_id=sender_id,
+            amount=amount,
+            cur_id=cur_id,
+            status=TransactionStatus.request,
+            ts=datetime.now(),
+        )
+
+    async def send(self, amount: int, cur_id: int, receiver_id: int) -> "Transaction":
+        b = await self.balance_xp(cur_id=cur_id)
+        assert b >= amount, "not enough balance"
+        transaction = await Transaction.create(
+            sender=self,
+            receiver_id=receiver_id,
+            amount=amount,
+            cur_id=cur_id,
+            validator_id=await self.get_val_id(amount, cur_id, receiver_id),
+            status=TransactionStatus.transfer,
+            ts=datetime.now(),
+        )
+        await transaction.sender_sign()
+        return transaction
+
+    async def send_by_req(self, req: "Transaction") -> "Transaction":
+        assert req.status == TransactionStatus.request, "its not a request"
+        assert not req.sender_id or self.id == req.sender_id, "sender id is incorrect"
+        b = await self.balance_xp(cur_id=req.cur_id)
+        assert b >= req.amount, "not enough balance"
+        sender_prv = Ed25519PrivateKey.from_private_bytes(self.prv)
+        req.sender = self
+        req.validator_id = await self.get_val_id(req.amount, req.cur_id, req.receiver_id)
+        req.status = TransactionStatus.transfer
+        req.ts = datetime.now()
+        req.proof = sender_prv.sign(req.pack())
+        await req.save()
+        return req
+        # async with in_transaction():
+        #     await req.fetch_related("sender", "validator")
+        #     req.proof = req.validator.proof_trans(req, sender_sig)
+        #     if not req.check():
+        #         raise ValueError("Transaction check failed")
+        #     req.status = TransactionStatus.valid
+        #     await req.save(update_fields=["proof", "status"])
+        # return transaction
 
     def name(self):
         return f"{self.first_name} {self.last_name}".rstrip()
@@ -300,6 +424,8 @@ class User(TgUser, TsTrait):
 
 @pre_save(User)
 async def person(_meta, user: User, _db, _updated: dict) -> None:
+    if not user.pub or not user.prv or len(user.prv + user.pub) != 64:
+        await user.keygen()
     if user.person_id:
         return
     user.person = await Person.create(name=f"{user.first_name} {user.last_name}".strip())
@@ -327,7 +453,7 @@ class Actor(Model):
     name: int = CharField(63)
     ex: ForeignKeyRelation[Ex] = ForeignKeyField("models.Ex", related_name="actors")
     ex_id: int
-    person: ForeignKeyNullableRelation[Person] = ForeignKeyField("models.Person", "actors", null=True)
+    person: ForeignKeyRelation[Person] = ForeignKeyField("models.Person", "actors")
     person_id: int
 
     agent: BackwardOneToOneRelation["Agent"]
@@ -393,7 +519,7 @@ class Synonym(BaseModel):
     txt: str = CharField(255, unique=True)
     boundary: Boundary = IntEnumField(Boundary, default=Boundary.no)
     curs: ManyToManyRelation[Cur] = ManyToManyField("models.Cur", "synonym_cur", related_name="synonyms")
-    val: int | None = SmallIntField(null=True)  # SynonymType dependent (e.g: no-slavic for name, approx for ppo..)
+    val: int | None = SmallIntField(null=True)  # SynonymType dependent (e.g., no-slavic for name, approx for ppo.)
     is_re: bool = BooleanField(default=False)
 
 
@@ -404,6 +530,9 @@ class CondSim(BaseModel):
     cond_rel: ForeignKeyRelation[Cond] = ForeignKeyField("models.Cond", "sims_rel")
     cond_rel_id: int  # old
 
+    class Meta:
+        table = "cond_sim"
+
 
 class PmGroup(Model):
     id: int = SmallIntField(True)
@@ -426,6 +555,7 @@ class Pm(Model):
     ok: bool = BooleanField(default=True)
     bank: bool | None = BooleanField(null=True)
     qr: bool = BooleanField(default=False)
+    fee: int | None = SmallIntField(null=True)  # /10_000
 
     typ: PmType | None = IntEnumField(PmType, null=True)
 
@@ -436,8 +566,8 @@ class Pm(Model):
     exs: ManyToManyRelation[Ex] = ManyToManyField("models.Ex", "pmex")  # no need. use pmexs[.exid]
     conds: BackwardFKRelation["CondParsed"]
     orders: BackwardFKRelation["Order"]
-    pmcurs: BackwardFKRelation["Pmcur"]  # no need. use curs
-    pmexs: BackwardFKRelation["Pmex"]
+    pmcurs: BackwardFKRelation["PmCur"]  # no need. use curs
+    pmexs: BackwardFKRelation["PmEx"]
 
     class Meta:
         table_description = "Payment methods"
@@ -466,25 +596,28 @@ class CondParsed(Model, TsTrait):
     slavic: bool = BooleanField(null=True)
     mtl_like: bool = BooleanField(null=True)
     scale: int = SmallIntField(null=True)
-    bank_side: bool = BooleanField(null=True)  # False - except this banks, True - only this banks
+    bank_side: bool = BooleanField(null=True)  # False - except these banks, True - only this banks
     sbp_strict: SbpStrict = IntEnumField(SbpStrict, default=SbpStrict.no)
     contact: str | None = CharField(127, null=True)
     done: bool = BooleanField(default=False, db_index=True)
     banks: ManyToManyRelation[Pm] = ManyToManyField("models.Pm", "cond_banks", related_name="conds")
 
+    class Meta:
+        table = "cond_parsed"
+
 
 class Ad(Model, TsTrait):
-    exid: int = BigIntField(unique=True)  # todo: спарить уникальность с биржей, тк на разных биржах могут совпасть
+    exid: int = BigIntField(db_index=True)  # todo: спарить уникальность с биржей, тк на разных биржах могут совпасть
     pair_side: ForeignKeyRelation[PairSide] = ForeignKeyField("models.PairSide", related_name="ads")
-    price: float = FloatField()
-    amount: float = FloatField()
-    min_fiat: float = FloatField()
-    max_fiat: float | None = FloatField(null=True)
-    auto_msg: str | None = CharField(255, null=True)
+    price: int = IntField()  # /10^cur.scale
+    premium: int = SmallIntField(null=True)  # /10_000
+    amount: int = IntField()  # /10^cur.scale
+    quantity: int = BigIntField(null=True)  # /10^coinex.scale
+    min_fiat: int = IntField()  # /10^cur.scale
+    max_fiat: int | None = IntField(null=True)  # /10^cur.scale
+    auto_msg: str | None = CharField(4095, null=True)
     status: AdStatus = IntEnumField(AdStatus, default=AdStatus.active)
 
-    ex: ForeignKeyRelation[Ex] = ForeignKeyField("models.Ex", related_name="ads")
-    ex_id: int
     cond: ForeignKeyNullableRelation[Cond] = ForeignKeyField("models.Cond", "ads", null=True)
     cond_id: int
     maker: ForeignKeyRelation[Actor] = ForeignKeyField("models.Actor", "my_ads")
@@ -499,6 +632,7 @@ class Ad(Model, TsTrait):
 
     class Meta:
         table_description = "P2P Advertisements"
+        unique_together = (("exid", "maker_id"),)
 
     # def epyds(self) -> tuple[PydModel, PydModel, PydModel, PydModel, PydModel, PydModel]:
     #     module_name = f"xync_client.{self.maker.ex.name}.pyd"
@@ -518,7 +652,7 @@ class MyAd(Model):  # Road
     ad_id: int  # new
     target_place: int = SmallIntField(default=1)
 
-    pay_req: ForeignKeyNullableRelation["PayReq"] = ForeignKeyField("models.PayReq", "my_ads", null=True)
+    pay_req: ForeignKeyNullableRelation["PayReq"] = ForeignKeyField("models.PayReq", "maked_ads", null=True)
     creds: ManyToManyRelation["Cred"] = ManyToManyField("models.CredEx", through="myad_cred", related_name="my_ads")
     race: BackwardFKRelation["Cred"]
 
@@ -526,27 +660,45 @@ class MyAd(Model):  # Road
         table = "my_ad"
 
 
-class Rival(Model, TsTrait):
-    actor: ForeignKeyRelation[Actor] = ForeignKeyField("models.Actor")
-    actor_id: int
-    exec_rate: int = SmallIntField()
-    completed: int = SmallIntField()
-
-
 class Race(Model):
     road: ForeignKeyRelation[MyAd] = ForeignKeyField("models.MyAd", "race")
     road_id: int
-    rival: ForeignKeyRelation[Rival] = ForeignKeyField("models.Rival", "races")
+    ceil: int = IntField()  # /10^cur.scale
+    target_place: int = SmallIntField(default=1)
+    vm_filter: bool = BooleanField(default=True)
+    started: bool = BooleanField(default=True)
+    updated_at: datetime | None = DatetimeSecField(auto_now=True)
+
+    stats: BackwardFKRelation["RaceStat"]
+
+    # def overprice_filter(self, ads: list[BaseAd]):
+    #     ads[0]
+    #     # вырезаем ads с ценами выше потолка
+    #     if ads and (self.ceil - Decimal(ads[0].price)) * k > 0:
+    #         if int(ads[0].userId) != self.actor.exid:
+    #             ads.pop(0)
+    #             self.overprice_filter(ads, self.ceil, k)
+
+
+class RaceStat(Model):
+    race: ForeignKeyRelation[Race] = ForeignKeyField("models.Race", "stats")
+    race_id: int
+    rivals: ManyToManyRelation[Ad] = ManyToManyField("models.Ad", "rivals", related_name="stats")
     rival_id: int
-    place: int = SmallIntField(default=1)
+    price: int = IntField()  # /10^cur.scale
+    premium: int = SmallIntField(null=True)  # /10_000
+    place: int = SmallIntField()
     created_at: datetime | None = DatetimeSecField(auto_now_add=True)
 
+    class Meta:
+        table = "race_stat"
+
 
 class PmRep(Model):
-    ex: ForeignKeyRelation[Ex] = ForeignKeyField("models.Ex", "pm_reps")
+    ex: ForeignKeyNullableRelation[Ex] = ForeignKeyField("models.Ex", "pm_reps", null=True)
     ex_id: int
     src: str | None = CharField(63, unique=True)
-    trgt: str | None = CharField(63)
+    target: str | None = CharField(63)
     used_at: datetime | None = DatetimeSecField(null=True)
 
     class Meta:
@@ -561,16 +713,18 @@ class PmAgent(Model):
     state: dict = JSONField(default={})
 
     class Meta:
+        table = "pm_agent"
         unique_together = (("pm_id", "user_id"),)
 
 
-class Pmcur(Model):  # for fiat with no exs tie
+class PmCur(Model):  # for fiat with no exs tie
     pm: ForeignKeyRelation[Pm] = ForeignKeyField("models.Pm")
     pm_id: int
     cur: ForeignKeyRelation[Cur] = ForeignKeyField("models.Cur")
     cur_id: int
 
     creds: BackwardFKRelation["Cred"]
+    sends: BackwardFKRelation["Transfer"]
     exs: ManyToManyRelation[Ex]
 
     class Meta:
@@ -582,7 +736,7 @@ class Pmcur(Model):  # for fiat with no exs tie
         include = "cur_id", "pm"
 
 
-class Pmex(BaseModel):  # existence pm in ex with no cur tie
+class PmEx(BaseModel):  # existence pm in ex with no cur tie
     ex: ForeignKeyRelation[Ex] = ForeignKeyField("models.Ex", "pmexs")
     ex_id: int
     pm: ForeignKeyRelation[Pm] = ForeignKeyField("models.Pm", "pmexs")
@@ -592,14 +746,14 @@ class Pmex(BaseModel):  # existence pm in ex with no cur tie
     exid: str = CharField(63)
     name: str = CharField(63)
 
-    banks: BackwardFKRelation["PmexBank"]
+    banks: BackwardFKRelation["PmExBank"]
 
     class Meta:
         unique_together = (("ex_id", "exid"),)  # , ("ex", "pm"), ("ex", "name")  # todo: tmp removed for HTX duplicates
 
 
-class PmexBank(BaseModel):  # banks for SBP
-    pmex: ForeignKeyRelation[Pmex] = ForeignKeyField("models.Pmex", "banks")
+class PmExBank(BaseModel):  # banks for SBP
+    pmex: ForeignKeyRelation[PmEx] = ForeignKeyField("models.PmEx", "banks")
     pmex_id: int
     exid: str = CharField(63)
     name: str = CharField(63)
@@ -607,11 +761,12 @@ class PmexBank(BaseModel):  # banks for SBP
     creds: ManyToManyRelation["Cred"]
 
     class Meta:
+        table = "pmex_bank"
         unique_together = (("pmex", "exid"),)
 
 
-# class Pmcurex(BaseModel):  # existence pm in ex for exact cur, with "blocked" flag
-#     pmcur: ForeignKeyRelation[Pmcur] = ForeignKeyField("models.Pmcur")
+# class PmCurEx(BaseModel):  # existence pm in ex for exact cur, with "blocked" flag
+#     pmcur: ForeignKeyRelation[PmCur] = ForeignKeyField("models.PmCur")
 #     pmcur_id: int
 #     ex: ForeignKeyRelation[Ex] = ForeignKeyField("models.Ex")
 #     ex_id: int
@@ -622,7 +777,7 @@ class PmexBank(BaseModel):  # banks for SBP
 
 
 class Cred(Model):
-    pmcur: ForeignKeyRelation[Pmcur] = ForeignKeyField("models.Pmcur")
+    pmcur: ForeignKeyRelation[PmCur] = ForeignKeyField("models.PmCur")
     pmcur_id: int
     detail: str = CharField(127)
     name: str | None = CharField(127, null=True)
@@ -631,9 +786,10 @@ class Cred(Model):
     person_id: int
     from_chat: int  # todo: ForeignKeyNullableRelation["Order"] = ForeignKeyField("models.Order", null=True)
 
-    banks: ManyToManyRelation[PmexBank] = ManyToManyField("models.PmexBank", related_name="creds")
+    banks: ManyToManyRelation[PmExBank] = ManyToManyField("models.PmExBank", related_name="creds")
 
     fiat: BackwardOneToOneRelation["Fiat"]
+    receives: BackwardOneToOneRelation["Transfer"]
     credexs: BackwardFKRelation["CredEx"]
     orders: BackwardFKRelation["Order"]
     pay_reqs: BackwardFKRelation["PayReq"]
@@ -667,9 +823,9 @@ class CredEx(Model):
 class Fiat(Model):
     cred: OneToOneRelation[Cred] = OneToOneField("models.Cred", "fiat")
     cred_id: int
-    amount: float = FloatField(default=0)
-    target: float = FloatField(default=0)
-    min_deposit: int = IntField(null=True)
+    amount: int = IntField()  # /10^cur.scale
+    target: int = IntField(null=True)  # /10^cur.scale
+    min_deposit: int = IntField(null=True)  # /10^cur.scale
 
     class Meta:
         table_description = "Currency balances"
@@ -687,13 +843,13 @@ class Fiat(Model):
 
 
 class Limit(Model):
-    pmcur: ForeignKeyRelation[Pmcur] = ForeignKeyField("models.Pmcur")
+    pmcur: ForeignKeyRelation[PmCur] = ForeignKeyField("models.PmCur")
     pmcur_id: int
     amount: int = IntField(null=True)  # '$' if unit >= 0 else 'transactions count'
     unit: int = IntField(default=30)  # positive: $/days, 0: $/transaction, negative: transactions count / days
     level: float | None = IntField(
         default=0, null=True
-    )  # 0 - same group, 1 - to parent group, 2 - to grandparent  # only for output trans, on input = None
+    )  # 0 - same group, 1 - to parent group, 2 - to grandparent # only for output trans, on input = None
     income: bool = BooleanField(default=False)
     added_by: ForeignKeyRelation["User"] = ForeignKeyField("models.User", related_name="limits")
     added_by_id: int
@@ -726,12 +882,11 @@ class Asset(Model):
     addr_id: int
     agent: ForeignKeyRelation[Agent] = ForeignKeyField("models.Agent", "assets")
     agent_id: int
-
     typ: AddrExType = IntEnumField(AddrExType, default=AddrExType.found)
-    free: float = FloatField()
-    freeze: float | None = FloatField(default=0)
-    lock: float | None = FloatField(default=0)
-    target: float | None = FloatField(default=0, null=True)
+    free: int = BigIntField()  # /10^coinex.scale
+    freeze: int | None = BigIntField(default=0)  # /10^coinex.scale
+    lock: int | None = BigIntField(default=0)  # /10^coinex.scale
+    target: int | None = BigIntField(default=0, null=True)  # /10^coinex.scale
 
     _name = {"asset__coin__ticker", "free"}
 
@@ -757,7 +912,8 @@ class PayReq(Model, TsTrait):
     parts: int = IntField(default=1)
     payed_at: datetime | None = DatetimeSecField(null=True)
 
-    my_ads: BackwardFKRelation["MyAd"]
+    maked_ads: BackwardFKRelation["MyAd"]
+    taken_orders: BackwardFKRelation["Order"]
 
     _icon = "pay"
     _name = {"ad_id"}
@@ -771,11 +927,14 @@ class Order(Model):
     exid: int = BigIntField(unique=True)  # todo: спарить уникальность с биржей, тк на разных биржах могут совпасть
     ad: ForeignKeyRelation[Ad] = ForeignKeyField("models.Ad", related_name="ads")
     ad_id: int
-    amount: float = FloatField()
+    amount: int = IntField()  # /10^cur.scale
+    quantity: int = BigIntField(null=True)  # /10^coinex.scale
     cred: ForeignKeyRelation[Cred] = ForeignKeyField("models.Cred", related_name="orders", null=True)
     cred_id: int | None
     taker: ForeignKeyRelation[Actor] = ForeignKeyField("models.Actor", "taken_orders")
     taker_id: int
+    payreq: ForeignKeyNullableRelation[PayReq] = ForeignKeyField("models.PayReq", "taken_orders", null=True)
+    payreq_id: int
     maker_topic: int = IntField(null=True)  # todo: remove nullability
     taker_topic: int = IntField(null=True)
     status: OrderStatus = IntEnumField(OrderStatus)
@@ -786,6 +945,7 @@ class Order(Model):
     chat_parsed: bool = BooleanField(default=False)
 
     msgs: BackwardFKRelation["Msg"]
+    transfer: BackwardOneToOneRelation["Transfer"]
 
     _name = {"cred__pmcur__pm__name"}
 
@@ -823,6 +983,91 @@ class Order(Model):
         exclude = ("taker", "ad", "cred", "msgs")
 
 
+class Transaction(Model):
+    amount: int = IntField()  # /10^cur.scale
+    cur: ForeignKeyRelation["Cur"] = ForeignKeyField("models.Cur", related_name="transfers")
+    cur_id: int
+    sender: ForeignKeyRelation[User] = ForeignKeyField("models.User", "sends", null=True)
+    sender_id: int
+    receiver: ForeignKeyRelation[User] = ForeignKeyField("models.User", "receives")
+    receiver_id: int
+    validator: ForeignKeyRelation[User] = ForeignKeyField("models.User", "validates", null=True)
+    validator_id: int
+    status: TransactionStatus = IntEnumField(TransactionStatus)
+    proof: bytes = UniqBinaryField(unique=True, null=True)  # len=128
+    ts: datetime | None = DatetimeSecField(null=True)
+
+    def is_expired(self) -> bool:
+        """Проверка истёк ли запрос"""
+        assert not self.proof and self.ts, "It is method for ttl requests only"
+        return datetime.now(timezone.utc) > self.ts
+
+    def is_valid_sender(self, sender_id: int) -> bool:
+        """Проверка может ли отправитель оплатить этот запрос"""
+        assert self.proof is None, "It is method for requests only"
+        return self.sender_id is None or self.sender_id == sender_id
+
+    def pack(self) -> bytes:
+        """Упаковка транзакции(или запроса, если нет sent_at) в 4+4+8+2+4=22 байт"""
+        return struct.pack(
+            ">LLQHL", self.sender_id, self.receiver_id, self.amount, self.cur_id, int(self.ts.timestamp())
+        )
+
+    async def sender_sign(self):
+        assert self.status == TransactionStatus.transfer, "It is method for transfers only, not requests"
+        await self.fetch_related("sender")
+        if (
+            not self.sender.pub or not self.sender.prv or not len(self.sender.pub + self.sender.prv) == 64
+        ):  # todo: remove check
+            await self.sender.keygen()
+        sender_prv = Ed25519PrivateKey.from_private_bytes(self.sender.prv)
+        self.proof = sender_prv.sign(self.pack())
+        await self.save(update_fields=["proof"])
+
+    async def vld_sign(self):
+        assert len(self.proof) == 64, "Sender did not signed yet"
+        await self.fetch_related("validator")
+        if (
+            not self.validator.pub or not self.validator.prv or not len(self.validator.pub + self.validator.prv) == 64
+        ):  # todo: remove check
+            await self.validator.keygen()
+        pk = Ed25519PrivateKey.from_private_bytes(self.validator.prv)
+        self.proof += pk.sign(self.pack() + self.proof)
+        self.status = TransactionStatus.valid
+        # Только две подписи: отправителя + бэкенда (128 байт)
+        await self.save(update_fields=["proof", "status"])
+
+    def check(self) -> bool:
+        """Проверка минимального доказательства"""
+        if len(self.proof) != 128:  # 64 + 64
+            return False  # wrong size
+        sender_sig, vld_sig = self.proof[:64], self.proof[64:128]
+        try:
+            # 1. Проверяем подпись отправителя
+            snd_pub = Ed25519PublicKey.from_public_bytes(self.sender.pub)
+            snd_pub.verify(sender_sig, (trans_packed := self.pack()))
+            # 2. Проверяем подпись бэкенда
+            vld_pub = Ed25519PublicKey.from_public_bytes(self.validator.pub)
+            vld_pub.verify(vld_sig, trans_packed + sender_sig)
+            return True
+        except InvalidSignature as _:
+            return False  # wring sign
+
+
+class Transfer(Model):
+    amount: int = IntField()  # /10^cur.scale
+    order: OneToOneNullableRelation[Order] = OneToOneField("models.Order", "transfer", null=True)
+    order_id: int
+    file: OneToOneNullableRelation["File"] = OneToOneField("models.File", "transfer", null=True)
+    file_id: int
+    pmid: int = BigIntField(null=True)
+    created_at: datetime | None = DatetimeSecField(auto_now_add=True)
+    accepted_at: datetime | None = DatetimeSecField(null=True)
+
+    class Meta:
+        unique_together = (("pmid", "order"),)
+
+
 class Msg(Model):
     tg_mid: int = IntField(null=True)
     txt: str = CharField(4095, null=True)
@@ -839,17 +1084,20 @@ class Msg(Model):
         exclude = ("receiver", "order")
 
     class Meta:
-        unique_together = (("order", "sent_at"),)
+        unique_together = (
+            ("order", "txt"),
+            ("order", "sent_at"),
+        )
 
 
 class Dep(Model, TsTrait):
     pid: str = CharField(31)  # product_id
-    apr: float = FloatField()
-    fee: float | None = FloatField(null=True)
+    apr: int = SmallIntField()  # /10_000
+    fee: int | None = SmallIntField(null=True)  # /10_000
     apr_is_fixed: bool = BooleanField(default=False)
     duration: int | None = SmallIntField(null=True)
     early_redeem: bool | None = BooleanField(null=True)
-    type_: DepType = IntEnumField(DepType)
+    typ: DepType = IntEnumField(DepType)
     # mb: renewable?
     min_limit: float = FloatField()
     max_limit: float | None = FloatField(null=True)
@@ -876,7 +1124,7 @@ class Dep(Model, TsTrait):
 
     class Meta:
         table_description = "Investment products"
-        unique_together = (("pid", "type_", "ex_id"),)
+        unique_together = (("pid", "typ", "ex_id"),)
 
 
 class Investment(Model, TsTrait):
@@ -910,6 +1158,7 @@ class ExStat(Model):
         return f"{self.ex_id} {self.action.name} {self.ok}"
 
     class Meta:
+        table = "ex_stat"
         table_description = "Ex Stats"
         unique_together = (("action", "ex_id"),)
 
@@ -920,7 +1169,7 @@ class ExStat(Model):
 class Vpn(Model):
     user: OneToOneRelation[User] = OneToOneField("models.User", related_name="vpn")
     user_id: int
-    priv: str = CharField(63, unique=True)
+    prv: str = CharField(63, unique=True)
     pub: str = CharField(63, unique=True)
     created_at: datetime | None = DatetimeSecField(auto_now_add=True)
 
@@ -935,9 +1184,6 @@ class Vpn(Model):
 
 
 class File(Model):
-    class UniqBinaryField(BinaryField):
-        indexable = True
-
     name: str = CharField(178, null=True, unique=True)
     typ: FileType = IntEnumField(FileType)
     ref: bytes = UniqBinaryField(unique=True)
@@ -945,7 +1191,8 @@ class File(Model):
     created_at: datetime | None = DatetimeSecField(auto_now_add=True)
 
     msg: BackwardOneToOneRelation[Msg]
-    pmex_logos: BackwardFKRelation[Pmex]
+    transfer: BackwardOneToOneRelation[Transfer]
+    pmex_logos: BackwardFKRelation[PmEx]
 
     _icon = "file"
     _name = {"name"}
